#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module datamodel for 3SE"""
from setuptools import setup


setup(name='datamodel',
      version='0.1',
      description='MTC datamodel',
      url='',
      author='',
      author_email='',
      license='',
      packages=['datamodel'],
      zip_safe=False)
